from astatsa.expectation_weighted.expectation_weighted import ExpectationWeighted

import numpy as np
from astatsa.utils.np_comparisons import assert_allclose

def test_weighted_1():
    ex = ExpectationWeighted()
    ex.update(np.array([1, 2]), np.array([0, 0]))    
    ex.update(np.array([1, 2]), np.array([0, 0]))
    assert np.all(np.isnan(ex.get_value()))

def test_weighted_2():
    ex = ExpectationWeighted()
    ex.update(np.array([1, 1]), np.array([0, 0]))
    ex.update(np.array([1, 2]), np.array([0, 0]))
    assert_allclose(42, ex.get_value(42))

def test_weighted_3():
    ex = ExpectationWeighted()
    ex.update(np.array([1, 1]), np.array([1, 1]))
    ex.update(np.array([2, 2]), np.array([0, 0]))
    assert_allclose([1, 1], ex.get_value())
    assert_allclose([1, 1], ex.get_mass())




def test_weighted_4():
    ex = ExpectationWeighted()
    ex.update(np.array([1, 1]), np.array([0, 101]))
    ex.update(np.array([2, 2]), np.array([23, 0]))
    assert_allclose([2, 1], ex.get_value())
    assert_allclose([23, 101], ex.get_mass())





import numpy as np
from contracts import contract
from ..expectation import Expectation

__all__ = ['MeanVariance']


# TODO: write tests for this

class MeanVariance:

    ''' Computes mean and variance of some stream. '''
    def __init__(self, max_window=None):
        self.Ex = Expectation(max_window)
        self.Edx2 = Expectation(max_window)
        self.num_samples = 0

    @contract(x='array', dt='float,>0')
    def update(self, x, dt=1.0):
        self.num_samples += dt
        self.Ex.update(x, dt)
        dx = x - self.Ex()
        dx2 = dx * dx
        self.Edx2.update(dx2, dt)

    def assert_some_data(self):
        if self.num_samples == 0:
            raise Exception('Never updated')

    def get_mean(self):
        return self.Ex()

    def get_var(self):
        return self.Edx2()

    def get_std_dev(self):
        return np.sqrt(self.get_var())
    
    def get_mean_stddev(self):
        """ returns a tuple (mean, stddev) """
        return self.get_mean(), self.get_std_dev()

    def publish(self, pub):
        if self.num_samples == 0:
            pub.text('warning',
                     'Cannot publish anything as I was never updated.')
            return

        pub.text('stats', 'Num samples: %s' % self.num_samples)

        mean = self.Ex()
        S = self.get_std_dev()

        if mean.ndim == 1:
            with pub.plot('mean') as pylab:
                pylab.plot(mean, 'k.')

            with pub.plot('std_dev') as pylab:
                pylab.plot(S, 'k.')
                a = pylab.axis()
                m = 0.1 * (a[3] - a[2])
                pylab.axis((a[0], a[1], 0, a[3] + m))
        else:
            pub.text('warning', 'Not implemented for ndim > 1')






from astatsa.utils.np_comparisons import assert_allclose
import numpy as np
from astatsa.expectation.expectation_fast import ExpectationFast

def check_my_doubt(exp_class):
    values = [100.0, 200.0]
    mean = 150.0
    
    for nd in [1, 10, 100, 1000, 2000]:
        
        samples = []
        for v in values:
            samples.extend([v] * nd)
        samples = np.array(samples)
    
        ex = exp_class()        
        for s in samples:
            ex.update(np.array([s]))
            
            # print('accum: %10.4f  mass: %10.4f ' % (ex.accum, ex.accum_mass))
            # ex.get_value()
        
            
        result = ex.get_value()
        # print('accum: %10.4f  mass: %10.4f ' % (ex.accum, ex.accum_mass))
                    
        assert_allclose(result, mean)
        
    

def test_fast():
    check_my_doubt(ExpectationFast)



import numpy as np 
from contracts import contract

from astatsa.expectation import Expectation
from astatsa.mean_variance import MeanVariance


__all__ = ['PredictionStats']


class  PredictionStats:

    @contract(label_a='str', label_b='str')
    def __init__(self, label_a='a', label_b='b'):
        self.label_a = label_a
        self.label_b = label_b
        self.Ea = MeanVariance()
        self.Eb = MeanVariance()
        self.Edadb = Expectation()
        self.R = None
        self.R_needs_update = True
        self.num_samples = 0
        self.last_a = None
        self.last_b = None

    @contract(a='array,shape(x)', b='array,shape(x)', dt='float,>0')
    def update(self, a, b, dt=1.0):
        self.Ea.update(a, dt)
        self.Eb.update(b, dt)
        da = a - self.Ea.get_mean()
        db = b - self.Eb.get_mean()
        self.Edadb.update(da * db, dt)
        self.num_samples += dt

        self.R_needs_update = True
        self.last_a = a
        self.last_b = b

    def get_correlation(self):
        ''' Returns the correlation between the two streams. '''
        if self.R_needs_update:
            std_a = self.Ea.get_std_dev()
            std_b = self.Eb.get_std_dev()
            p = std_a * std_b
            zeros = p == 0
            p[zeros] = 1
            R = self.Edadb() / p
            R[zeros] = np.NAN
            self.R = R
        self.R_needs_update = False
        return self.R



    



from astatsa.expectation import ExpectationFast, ExpectationSlow
import numpy as np
from astatsa.utils import assert_allclose
from .test_doubt import check_my_doubt
from astatsa.expectation import ExpectationFaster


def test_efficient_exp():
    shape = (4, 4)
    for exp_class in [ExpectationSlow, ExpectationFast, ExpectationFaster]:
        yield check_expectation_one, exp_class

        sequence = (np.random.randn(*shape) for i in range(1))
        yield check_efficient_exp, exp_class, sequence

        sequence = (np.random.randn(*shape) for i in range(5))
        yield check_efficient_exp, exp_class, sequence

        sequence = (np.random.randn(*shape) for i in range(1000))
        yield check_efficient_exp, exp_class, sequence

        yield check_my_doubt, exp_class


def check_expectation_one(exp_class):
    e = exp_class()
    x = np.random.rand(2, 2)
    dt = np.random.rand()
    e.update(x, dt)
    v = e.get_value()
    assert_allclose(v, x)


def check_efficient_exp(exp_class, sequence):
    """ Checks that the dt is working ok. """
    exp1 = exp_class()
    xs = []
    T = 0
    for x in sequence:
        dt = np.random.rand()
        xs.append(x * dt)
        T += dt
        exp1.update(x, dt=dt)

    es1 = exp1.get_value()
    expected = np.array(xs).sum(axis=0) / T

    assert_allclose(es1, expected)







__all__ = ['Expectation', 'ExpectationSlow', 'ExpectationFast']

from contracts import contract
import numpy as np

from .interface import ExpectationInterface
from .expectation_fast import ExpectationFast, ExpectationFaster
from .expectation_slow import ExpectationSlow 

Expectation = ExpectationFast



''' Some utils for checking and comparing arrays. '''

__all__ = ['check_all_finite', 'all_finite', 'assert_allclose', 'check_allclose']

from contracts import contract
from numpy.ma.core import allclose
import numpy as np


try:
    from numpy.testing.utils import assert_allclose  # @UnusedImport
except ImportError:
    def assert_allclose(actual, desired, rtol=1e-7, atol=0,
                        err_msg='', verbose=True):
        ''' Backporting assert_allclose from Numpy 1.5 to 1.4 '''
        from numpy.testing.utils import assert_array_compare  # @UnresolvedImport
    
        def compare(x, y):
            return np.allclose(x, y, rtol=rtol, atol=atol)
    
        actual, desired = np.asanyarray(actual), np.asanyarray(desired)
        header = 'Not equal to tolerance rtol=%g, atol=%g' % (rtol, atol)
        assert_array_compare(compare, actual, desired, err_msg=str(err_msg),
                             verbose=verbose, header=header)


def check_allclose(actual, desired, rtol=1e-7, atol=0,
                        err_msg='', verbose=True):
    if not allclose(actual, desired, rtol=rtol, atol=atol):
        err_msg += '\n%s != %s ' % (actual, desired)
        raise Exception(err_msg)




def assert_allequal_verbose(a, b, **kwargs):  # XXX @UnusedVariable
    failures = a != b
    if np.any(failures):
        error = show_differences(a, b, failures, "a==b")
        raise ValueError(error)


def assert_allclose_verbose(a, b, **kwargs):
    try:
        assert_allclose(a, b, **kwargs)
    except AssertionError as e:
        is_failure = a != b
        condition = 'a==b'  # XXX
        error = show_differences(a, b, is_failure, condition)
        raise Exception('%s\n\n%s' % (e, error))

#
# def close(a, b, rtol=1.e-5, atol=1.e-8):
#    """
#    Same as allclose() but returns the result element by element.
#    """
#    x = array(a, copy=False, ndmin=1)
#    y = array(b, copy=False, ndmin=1)
#    xinf = isinf(x)
#    if not all(xinf == isinf(y)):
#        return False
#    if not any(xinf):
#        return all(less_equal(absolute(x-y), atol + rtol * absolute(y)))
#    if not all(x[xinf] == y[xinf]):
#        return False
#    x = x[~xinf]
#    y = y[~xinf]
#    return all(less_equal(absolute(x-y), atol + rtol * absolute(y)))


@contract(a='shape(x)',
          b='shape(x)',
          is_failure='shape(x)',  # ,array(bool)',
          condition='str',
          MAX_N='int,>=1')
def show_differences(a, b, is_failure, condition, MAX_N=4):
    """ Returns a string summarizing the failures. """
    some, = np.nonzero(is_failure.flat)
    num = a.size
    num_fail = len(some)
    perc = 100.0 * num_fail / num
    error = ("In this array, %d/%d (%f%%) of elements do not respect "
             "the condition %s." % (num_fail, num, perc, condition))

    N = min(len(some), MAX_N)
    error += '\nThese are the first %d:' % N
    for i in range(min(len(some), MAX_N)):
        j = some[i]
        error += ('\n a.flat[%d] = %10s  b.flat[%d] = %10s ' % (j, a.flat[j],
                                                                j, b.flat[j]))
    return error


@contract(a='shape(x)',
          is_failure='shape(x)',  # ,array(bool)',
          condition='str',
          MAX_N='int,>=1')
def show_some(a, is_failure, condition, MAX_N=4):
    """ Returns a string summarizing the failures. """
    some, = np.nonzero(is_failure.flat)
    num = a.size
    num_fail = len(some)
    perc = 100.0 * num_fail / num
    error = ("In this array, %d/%d (%f%%) of elements do not respect "
             "the condition %s." % (num_fail, num, perc, condition))

    N = min(len(some), MAX_N)
    error += '\nFirst %d of %d elements:' % (N, num)
    for i in range(min(len(some), MAX_N)):
        j = some[i]
        error += ('\n a[%d] = %10s ' % (j, a.flat[j]))
    return error


def check_all_finite(value):
    """ Raises a ValueError if some values are not finite. """
    if all_finite(value):
        return

    invalid = np.logical_not(np.isfinite(value))
    msg = show_some(value, invalid, 'finite')

    msg += '\n%s %s ' % (np.min(value), np.max(value))
    raise ValueError(msg)


def all_finite(x):
    """ Fast way to check that all elements of an array are finite. """
    return np.isfinite(np.min(x)) and np.isfinite(np.max(x))




from . import ExpectationInterface, contract

__all__ = ['ExpectationSlow']


class ExpectationSlow(ExpectationInterface):
    
    ''' A class to compute the mean of a quantity over time '''
    def __init__(self, max_window=None):
        ''' 
            If max_window is given, the covariance is computed
            over a certain interval. 
        '''
        self.num_samples = 0.0
        self.value = None
        self.max_window = max_window

    @contract(value='array', dt='>0')
    def update(self, value, dt=1.0):
        if self.value is None:
            self.value = value
        else:
            self.value = weighted_average(self.value, float(self.num_samples),
                                          value, float(dt))
        self.num_samples += dt
        if self.max_window and self.num_samples > self.max_window:
            self.num_samples = self.max_window

    def get_value(self):
        return self.value

    def get_mass(self):
        return self.num_samples


@contract(A='array', wA='>=0', B='array', wB='>=0')
def weighted_average(A, wA, B, wB):
    mA = wA / (wA + wB)
    mB = wB / (wA + wB)
    return (mA * A + mB * B)



__version__ = '1.0'

import logging
logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

from .expectation import *
from .expectation_weighted import *
from .mean_variance import *
from .prediction import *



from abc import ABCMeta, abstractmethod
from contracts import contract

class ExpectationInterface():
    __metaclass__ = ABCMeta
    
    
    @abstractmethod
    @contract(value='array', dt='float,>=0', returns='None')
    def update(self, value, dt=1.0):
        pass
    

    @abstractmethod
    @contract(returns='array')
    def get_value(self):
        pass



from . import ExpectationInterface, contract, np
from astatsa.utils import check_all_finite
import warnings

__all__ = ['ExpectationFast', 'ExpectationFaster']


class ExpectationFast(ExpectationInterface):
    ''' A more efficient implementation. '''

    MAX_MASS = 100  # TODO: watch np.max(self.accum) instead

    def __init__(self, max_window=None, extremely_fast=False):
        '''  
            extremely_fast: saves memory; might crash on some python 
            interpreters
            TODO: put automatic tests to detect this
        '''
        self.max_window = max_window
        self.accum_mass = 0.0
        self.accum = None
        self.needs_normalization = True
        self.extremely_fast = extremely_fast
        
    def merge(self, other):
        warnings.warn('to test')
        assert isinstance(other, ExpectationFast)
        self.update(other.get_value(), other.accum_mass)

    @contract(cur_mass='float,>=0')
    def reset(self, cur_mass=1.0):
        self.accum = self.get_value()
        self.accum_mass = cur_mass

    @contract(value='array', dt='float,>=0')
    def update(self, value, dt=1.0):
        check_all_finite(value)

        if self.accum is None:
            self.accum = value * dt
            self.accum_mass = dt
            self.needs_normalization = True
            self.buf = np.empty_like(value)
            self.buf.fill(np.NaN)
            self.result = np.empty_like(value)
            self.result.fill(np.NaN)
        else:
            if self.extremely_fast:
                np.multiply(value, dt, self.buf)  # buf = value * dt
                np.add(self.buf, self.accum, self.accum)  # accum += buf
            else:
                self.buf = value * dt
                self.accum += self.buf

            self.needs_normalization = True
            self.accum_mass += dt

        if self.max_window and self.accum_mass > self.max_window:
            self.accum = self.max_window * self.get_value()
            self.accum_mass = self.max_window

        
        # Do not let pass too much before normalization
        if self.accum_mass > ExpectationFast.MAX_MASS:
            self.get_value()

    def get_value(self):
        if self.accum is None:
            raise ValueError('No value given yet.')
        if self.needs_normalization:
            # In the case dt=0 for the first sample
            if self.accum_mass > 0:
                ratio = 1.0 / self.accum_mass
            else:
                ratio = 1.0
            if self.extremely_fast:
                np.multiply(ratio, self.accum, self.result)
            else:
                self.result = ratio * self.accum
            self.needs_normalization = False
        return self.result

    def __call__(self):
        return self.get_value()

    def get_mass(self):
        return self.accum_mass


class ExpectationFaster(ExpectationFast):
    def __init__(self):
        ExpectationFast.__init__(self, extremely_fast=True)
        
        
        



from abc import ABCMeta, abstractmethod
from contracts import contract

class ExpectationWeightedInterface():
    __metaclass__ = ABCMeta
    
    
    @abstractmethod
    @contract(value='array,shape(x)', weight='array(>=0),shape(x)', returns='None')
    def update(self, value, weight):
        """ 
            weight: weight for each single element. 
            if 0, ignore this value
        """
        pass

    @abstractmethod
    @contract(returns='array')
    def get_value(self):
        """ 
            Returns the value of the expectation. 
            Raises ValueError if never updated.
            
            For those weights which were never updated, return np.nan. 
        """
        pass



from astatsa.expectation_weighted.interface import ExpectationWeightedInterface
from astatsa.utils.np_comparisons import check_all_finite
from contracts import contract
import numpy as np
import warnings
 
 
class ExpectationWeighted(ExpectationWeightedInterface):
    ''' 
        This operator allows, for each time step, to give a different weight
        to each element. The weight tensor should have the same shape as the value.
     '''
 
    def __init__(self):
        self.mass = None
        self.accum = None
         
        self._result = None
         
    def merge(self, other):
        warnings.warn('To test')
        assert isinstance(other, ExpectationWeighted)
        self.update(other.get_value(), other.get_mass())
        
        
    @contract(value='array,shape(x)', weight='array(>=0),shape(x)')
    def update(self, value, weight):
        check_all_finite(value)
        check_all_finite(weight)
        assert value.shape == weight.shape
         
        # If first time
        if self.accum is None:
            self.accum = (value * weight).astype('float64')
            self.mass = weight.astype('float64')
            self._result = None
        else:
            self.accum += value * weight
            self.mass += weight
            self._result = None
 
    def get_value(self, fill_value=np.nan):
        """ Returns the value of the expectation. Raises ValueError if never updated. """
        if self.accum is None:
            msg = 'No value given yet.'
            raise ValueError(msg)
 
        if self._result is not None:
            return self._result
        
        self._result = self._compute_value(fill_value)
        return self._result
    
    def _compute_value(self, fill_value):
        zeros = self.mass == 0
        mass = self.mass.copy()
        # set to one the mass
        mass[zeros] = 1
        result = self.accum / mass
        result[zeros] = fill_value 
        return result.copy()
 
    def get_mass(self):
        return self.mass.copy()
